package KGS::Game::Board;

use KGS::Constants;

sub new {
   my $class = shift;
   my $size = shift;
   bless {
         max      => $size - 1,
         board    => [map [(0) x $size], 1 .. $size],
         captures => [0, 0], # captures
         time     => [],
         comment  => "",
      },
      $class;
}

# inefficient and primitive, I hear you say?
# well... you are right :)
# use an extremely dumb floodfill algorithm to get rid of captured stones
sub capture {
   my ($self, $mark, $x, $y) = @_;

   my %seen;
   my @found;
   my @nodes = ([$x,$y]);

   while (@nodes) {
      my ($x, $y) = @{pop @nodes};
      unless ($seen{"$x,$y"}++) {
         if ($self->{board}[$x][$y] & $mark) {
            push @found, [$x, $y];

            push @nodes, [$x-1, $y] if $x > 0;
            push @nodes, [$x+1, $y] if $x < $self->{max};
            push @nodes, [$x, $y-1] if $y > 0;
            push @nodes, [$x, $y+1] if $y < $self->{max};
         } elsif (!($self->{board}[$x][$y] & (MARK_B | MARK_W))) {
            return;
         }
      }
   }

   $self->{captures}[ $mark == MARK_B ? WHITE : BLACK ] += @found;

   # capture!
   for (@found) {
      my ($x, $y) = @$_;
      $self->{board}[$x][$y] &= ~(MARK_B | MARK_W | MARK_MOVE);
   }
}

sub capture4 {
   my ($self, $mark, $x, $y) = @_;

   $self->capture($mark, $x-1, $y) if $x > 0            && $self->{board}[$x-1][$y] & $mark;
   $self->capture($mark, $x+1, $y) if $x < $self->{max} && $self->{board}[$x+1][$y] & $mark;
   $self->capture($mark, $x, $y-1) if $y > 0            && $self->{board}[$x][$y-1] & $mark;
   $self->capture($mark, $x, $y+1) if $y < $self->{max} && $self->{board}[$x][$y+1] & $mark;
}

sub interpret_path {
   my ($self, $path) = @_;

   my $move;

   for (@$path) {
      # mask out all labeling except in the last node
      my $nodemask = ~(
            $_ == $path->[-1]
               ? 0
               : MARK_SQUARE | MARK_TRIANGLE | MARK_CIRCLE | MARK_LABEL
         );
               
      while (my ($k, $v) = each %$_) {
         if ($k =~ /^(\d+),(\d+)$/) {
            $self->{board}[$1][$2] =
               $self->{board}[$1][$2]
               & ~$v->[1]
               | $v->[0]
               & $nodemask;

            $self->capture4(MARK_W, $1, $2) unless $v->[0] & (MARK_B | MARK_MOVE) ^ (MARK_B | MARK_MOVE);
            $self->capture4(MARK_B, $1, $2) unless $v->[0] & (MARK_W | MARK_MOVE) ^ (MARK_W | MARK_MOVE);
         } elsif ($k eq "time") {
            $self->{time}[0] = $v->[0] if defined $v->[0];
            $self->{time}[1] = $v->[1] if defined $v->[1];
         } elsif ($k eq "comment") {
            $self->{comment} .= "Move $move\n$v";
         }
      }

      $move++;
   }
}

1;

