package KGS::Game::Tree;

use KGS::Constants;

# _should_ be unified, but the two types are slightly different
# (this really looks like a grown, as opposed to designed, protocol)
# despite the claim of "innovation" etc...
my %marker_type = (
   triangle => MARK_TRIANGLE,
   square   => MARK_SQUARE,
   circle   => MARK_CIRCLE,
   small_b  => MARK_SMALL_B,
   small_w  => MARK_SMALL_W,
);

# exclusion masks... the bit on the left excludes (removes) the ones on the right
my %exclude_type = (
   &MARK_TRIANGLE => MARK_SQUARE | MARK_TRIANGLE | MARK_CIRCLE | MARK_LABEL,
   &MARK_SQUARE   => MARK_SQUARE | MARK_TRIANGLE | MARK_CIRCLE | MARK_LABEL,
   &MARK_CIRCLE   => MARK_SQUARE | MARK_TRIANGLE | MARK_CIRCLE | MARK_LABEL,
   &MARK_LABEL    => MARK_SQUARE | MARK_TRIANGLE | MARK_CIRCLE | MARK_LABEL,
   &MARK_SMALL_B  => MARK_SMALL_B | MARK_SMALL_W | MARK_B | MARK_W | MARK_MOVE,
   &MARK_SMALL_W  => MARK_SMALL_B | MARK_SMALL_W | MARK_B | MARK_W | MARK_MOVE,
   &MARK_GRAY_B   =>                               MARK_B | MARK_W | MARK_MOVE | MARK_GRAY_B | MARK_GRAY_W,
   &MARK_GRAY_W   =>                               MARK_B | MARK_W | MARK_MOVE | MARK_GRAY_B | MARK_GRAY_W,
   &MARK_B        => MARK_SMALL_B | MARK_SMALL_W | MARK_B | MARK_W | MARK_MOVE | MARK_GRAY_B | MARK_GRAY_W,
   &MARK_W        => MARK_SMALL_B | MARK_SMALL_W | MARK_B | MARK_W | MARK_MOVE | MARK_GRAY_B | MARK_GRAY_W,
);

sub init_tree {
   my ($self) = @_;
   $self->{tree} = [ { id => 0 } ];
}

sub update_tree {
   my ($self, $tree) = @_;

   my $node = $self->{curnode};

   for (@$tree)  {
      my ($type, @arg) = @$_;
      if ($type eq "add_child") {
         $node = $self->{tree}[$arg[0] + 1]
            or die "FATAL: referencing nonexistent node $arg[0]+1!";

         my $new = { parent => $node->{id}, id => scalar @{$self->{tree}} };

         push @{$self->{tree}}, $new;
         push @{$node->{children}}, $new->{id};

         $node = $new;

      } elsif (my $bit = $marker_type{$type}) {
         my $ref = $node->{"$arg[1],$arg[2]"} ||= [];

         $ref->[0] &= ~$exclude_type{$bit};
         $ref->[0] |= $bit if $arg[0] != 2;
         $ref->[1] |= $exclude_type{$bit};

      } elsif ($type eq "label") {
         my $ref = $node->{"$arg[1],$arg[2]"} ||= [];

         $ref->[0] &= ~$exclude_type{&MARK_LABEL};
         undef $ref->[2];

         if (length $arg[0]) {
            $ref->[0] |= MARK_LABEL;
            $ref->[1] |= $exclude_type{&MARK_LABEL};
            $ref->[2] = $arg[0];
         }

      } elsif ($type eq "move" or $type eq "addstone" or $type eq "gray") {
         my $ref = $node->{"$arg[1],$arg[2]"} ||= [];

         my $bit = $arg[0] == 0 ? $bit | ($type eq "gray" ? MARK_GRAY_B : MARK_B)
                 : $arg[0] == 1 ? $bit | ($type eq "gray" ? MARK_GRAY_W : MARK_W)
                                : 0;

         $ref->[0] &= ~$exclude_type{$bit};
         $ref->[0] |= $bit | ($type eq "move" ? MARK_MOVE : 0);
         $ref->[1] |= $exclude_type{$bit};

         $self->event_move if $self->{loaded} & $type eq "move";

      } elsif ($type eq "comment") {
         $node->{comment} .= $arg[0];

      } elsif ($type eq "player") {
         $node->{player}[$arg[0]] = $arg[1];

      } elsif ($type eq "time_left") {
         $node->{time}[$arg[0]] = [$arg[1], $arg[2]];

      } elsif ($type eq "more") {
         $self->{more} = 1;
         $self->msg(get_tree => channel => $self->{channel}, node => @{$self->{tree}} - 1);

      } elsif ($type eq "done") {
         $self->{loaded} = 1;
         # nop

      } else {
         warn "unknown node parameter $type (@arg), ignored\n";
      }
   }

   $self->{curnode} = $node;
}

sub get_path {
   my ($self) = @_;

   my @nodes;

   my $node = $self->{curnode};

   for(;;) {
      push @nodes, $node;
      last unless $node->{parent};
      $node = $self->{tree}[$node->{parent}];
   }

   [reverse @nodes];
}

sub set_node {
   my ($self, $node) = @_;

   $self->{curnode} = $self->{tree}[$node + 1]
      or die "set_node to undefined tree node $node+1";
}

sub event_move { }
sub event_update_tree { }

1;

