package KGS::Listener;

=item new

=cut

sub new {
   my $class = shift;
   bless { @_ }, $class;
}

=cut

=item $listener->listen($conn, [msgtype...])

Registers the object to receive callback messages of the named type(s). No
msgtypes means all msgtypes.

=item $listener->unlisten

Unregisters the object again.

=cut

sub listen {
   my ($self, $conn, @types) = @_;

   if ($conn) {
      $self->unlisten;
      $self->{conn} = $conn;
      $self->{listen_types} = \@types;
      $conn->register($self, @types);
   }
}

sub unlisten {
   my ($self) = @_;

   (delete $self->{conn})
      ->unregister($self, @{$self->{listen_types}}) if $self->{conn};
}

=item $listener->msg($type, %args);

Calls the C<msg> method of the connection when in listen state.

=cut

sub msg {
   my $self = shift;
   $self->{conn}->msg(@_) if $self->{conn};
}

sub inject {
   my ($self, $msg) = @_;

   if (my $cb = $self->can("inject_$msg->{type}")) {
      $cb->($self, $msg);
   } else {
      $self->inject_fallback ($msg);
   }
}

sub inject_fallback {
}

1;

