package KGS::Protocol;

use Carp;
use Compress::Zlib;
use Time::HiRes;

my $KGSHOST = "kgs.kiseido.com";
my $KGSPORT = 2379;
my $SGFHOST = "216.93.172.124";

our $NOW; # the time the last packet was received

sub MSG_CHANNEL() { 0x4000 }

# thank god (ok, I am an atheist) that it was soo simple...
sub hash_pw {
   require Math::BigInt; # I insist on 32-bit-perl.. should use C
   # $hash must be 64 bit
   my $hash = new Math::BigInt;
   warn $hash = $hash * 1055 + ord for split //, $_[0];
   $hash;
}

=item new arg => value,...

=cut

sub new {
   my $class = shift;
   my $self = bless {
      @_,
   }, $class;
   $self;
}

my @recv_rules = (
      [ruleset  => \&gb],
      [size     => \&gb],
      [handicap => \&gb],
      [komi     => \&gw2],
      [timesys  => \&gb],
      [time     => \&gi],
      [interval => \&gi], # byo-yomi time / canadian time
      [count    => \&gw], # periods / moves
   );

# decode some binary stream using a template
sub decode_template {
   my ($template, $data) = @_;
   my $msg = {};
   for (@{$template || []}) {
      if (ref $_) {
         my ($name, $cvt, $guard) = @$_;
         $guard ||= sub { !exists $_[0]{$name} };
         if (ref $cvt) {
            while ($guard->($msg) && length $data) {
               last unless length $data;
               (my ($dec), $data) = ARRAY eq ref $cvt
                                       ? decode_template($cvt, $data)
                                       : $cvt->($data);
               if ($guard == \&guard_repeat) {
                  push @{$msg->{$name}}, $dec;
               } else {
                  $msg->{$name} = $dec;
               }
            }
         } else {
            $msg->{$name} = $cvt;
         }
      } else {
         bless $msg, $_;
      }
   }
   ($msg, $data);
}

# low-level serialising/deserializing
# pl 64 bit
# pi 32 bit
# pw 16 bit
# pa zero-padded ascii
# pf flag (boolean)

sub pl($)  {
   pack "VV",
      $_[0]         & 0xffffffff,
      ($_[0] >> 32) & 0xffffffff;
}

sub pi      { pack "V", $_[0] }
sub pw      { pack "v", $_[0] }
sub pa      { pack "a$_[1]", $_[0] }
sub pb      { pack "c", $_[0] }
sub pf      { pack "c", ! ! $_[0] }

sub pa5     { pa $_[0], 5 }
sub pa12    { pa $_[0], 12 }

sub pz      { pack "v*", map ord, split //, $_[0] }

sub pany    { $_[0] }

sub gi      { 
   my ($r, $data) = unpack "V a*", $_[0];
   ((unpack "i", pack "I", $r), $data);
}
sub gl      { 
   my ($lo, $hi, $r) = unpack "VV a*", $_[0];
   ($lo + $hi * 2**32, $r);
}
sub gw      { unpack "v a*", $_[0] }
sub gb      { unpack "C a*", $_[0] }
sub ga      { unpack "Z$_[1] a*", $_[0] }
sub ga12    { ga $_[0], 12 }
sub gany    { ($_[0], "") }
sub gf      { &gb } # flag

sub gs      {
   my ($s, $r) = unpack "a".(2*$_[1])." a*", $_[0];
   ((join "", map chr, grep $_, unpack "v*", $s), $r);
}

sub gs50    { gs $_[0], 50 }

sub gw2  { # gi / 2
   my ($r, $data) = &gw;
   ($r * 0.5, $data);
}

sub gi2  { # gi / 2
   my ($r, $data) = &gi;
   ($r * 0.5, $data);
}

sub gi4  { # gi / 4
   my ($r, $data) = &gi;
   ($r * 0.25, $data);
}

# zero or message-terminated
sub gz {
   my ($data) = @_;
   my $c;
   $r = "";
   while (length $data) {
      ($c, $data) = unpack "v a*", $data;
      $c == 0 and last;
      $r .= chr $c;
   }
   ($r, $data);
}

#
# this was the most horrible thing to decode. still not everything is decoded correctly(?)
sub gtree {
   my $data = $_[0];
   my @r;
   while (length $data) {
      ((my $type), $data) = gb $data;

      if ($type == 255) {
         ((my $node), $data) = gi $data;
         push @r, [add_child => $node];

      } elsif ($type == 254) {
         push @r, ["done"];

      } elsif ($type == 253) {
         push @r, ["type253"];
         # ????

      } elsif ($type == 252) { # even less clear
         push @r, ["more"];

      } elsif ($type == 10) {
         # as usual, wms finds yet another way to duplicate code... oh well, what a mess.
         # (no wonder he is so keen on keeping it secret...)
         ((my $param), $data) = decode_template \@recv_rules, $data;

         push @r, [params => $param];

      } elsif ($type == 9 || $type == 22 || $type == 25 || $type == 31 || $type == 32) {
         ((my $arg), $data) = gz $data;

         push @r, [({
               9  => "comment",
               22 => "unknown_comment22",
               25 => "copyright", #?
               31 => "date",
               32 => "unknown_comment32",
               })->{$type} => $arg];

      } elsif ($type == 11 || $type == 12) {
         ((my $a1), $data) = gz $data;
         push @r, [player => $type - 11, $a1];

      } elsif ($type == 13 || $type == 14) {
         ((my $arg), $data) = gi $data;

         push @r, [rank => $type - 13, $arg];

      } elsif ($type == 15 || $type == 16) {
         ((my $a1), $data) = gi $data;
         ((my $a2), $data) = gw $data;

         push @r, [time_left => $type - 15, $a1 * 0.001, $a2];

      } elsif ($type == 17 || $type == 18) {
         ((my $a1), $data) = gw $data;

         push @r, [score => $type - 17, $a1 * 0.25];

      } elsif ($type == 19) {
         ((my $a1), $data) = gi2 $data;

         push @r, [result => $a1];

      } elsif ($type == 30) {
         ((my $a1), $data) = gb $data;

         push @r, [active_player => $a1];

      } elsif ($type == 0) { # label(?)
         ((my $c1), $data) = gw $data;
         ((my $c2), $data) = gw $data;
         ((my $c3), $data) = gw $data;
         ((my $a1), $data) = gb $data; # x
         ((my $a2), $data) = gb $data; # y

         my $label;
         $label .= chr $c1 if $c1;
         $label .= chr $c2 if $c2;
         $label .= chr $c3 if $c3;

         # empty label == remove label
         push @r, [label => $label, $a1, $a2];

      } elsif ($type > 0 && $type < 9) {
         ((my $a1), $data) = gb $data; # 255 == pass
         ((my $a2), $data) = gb $data; # x
         ((my $a3), $data) = gb $data; # y(?)

         # 1 marker type triangle
         # 2 marker type square
         # 3 marker type circle
         # 4 small stone b
         # 5 small stone w
         # 6 grayed out
         # 7 move
         # 8 also move(?) or preset?
         #
         # $a1 is probably player again (2 == remove)
         # x is from left 0 to right boardsize-1
         # y is from top  0 to bottom boardsize-1
         push @r, [(qw(triangle square circle small_b small_w gray move addstone))[$type-1] => $a1, $a2, $a3];

      } else {
         push @r, [unknown => $type];
         last;
      }
   }
   (\@r, "");
}

my $sendmsgs;
my $recvmsgs;

{
   sub gen_send {
      my ($name, $type, @args) = @_;
      $sendmsgs->{$name} = [
         [_type => \&pw, $type],
         ($type & MSG_CHANNEL ? [channel => \&pw] : ()),
         @args,
      ];
   }

   sub gen_recv {
      my ($name, $type, @args) = @_;
      $recvmsgs->{$type} = [
         [type => $name],
         ($type & MSG_CHANNEL ? [channel => \&gw] : ()),
         @args,
      ];
   }

   my @send_user = ([name => \&pa12], [flags => \&pi, 1]);

   gen_send login       => 0x0000,
            [ver_major => \&pi, 2], [ver_minor => \&pi, 4], [ver_micro => \&pi, 67],
            [login     => \&pa12],
            [password  => \&pl, 0],
            [guest     => \&pf, 1],
            [_unknown3 => \&pw, 0],
            [locale    => \&pa5, "en_US"],
            [javavm    => \&pany, "1.4.1_01-b01:Swing app:Sun Microsystems Inc."];

   gen_send server_stats => 0x0014;

   gen_send list_rooms  => 0x0318, [group => \&pb];
   gen_send new_room    => 0x031a, [name => \&pa12],
                                   [i1 => \&pi, 0],
                                   [b1 => \&pb, 0x00],
                                   [b2 => \&pb, 0xff],
                                   [b3 => \&pb, 0xff],
                                   [group => \&pb, 0x01],
                                   [name => \&pz],
                                   [description => \&pz],
                                   [flags => \&pb]; # 0x10 private

   gen_send join_room   => 0x4300, @send_user;
   gen_send msg_room    => 0x4301, [name => \&pa12], [message => \&pz];
   gen_send part_room   => 0x4302, [name => \&pa12];
   gen_send desc_room   => 0x4319; # fetch room description

   gen_send send_chal   => 0x4400, [black => \&pa12], [white => \&pa12]; # @send_challenge
   gen_send join_game   => 0x4403, @send_user;
   gen_send part_game   => 0x4404, [name => \&pa12];
   gen_send set_tree    => 0x4405, [tree => \&ptree];#d# no ptree
   gen_send get_tree    => 0x4408, [node => \&pi];
   gen_send claim_win   => 0x440c, [_byte => \&pb ];
   gen_send add_time    => 0x440d, [time => \&pi], [player => \&pb];
   gen_send grant_undo  => 0x440f;
   gen_send resign_game => 0x4410, [player => \&pb];
   gen_send set_teacher => 0x441a, [name => \&pa12];
   gen_send add_user    => 0x4422, [othername => \&pa12], [name => \&pa12]; # gives user access
   gen_send set_privacy => 0x4423, [private => \&pb];
   gen_send reject_chal => 0x4429, [name => \&pa12];
   gen_send req_result  => 0x4433;

   my @recv_user       = (KGS::User::, [name => \&ga12], [flags => \&gi]);
   my @recv_chal_def   = ([gametype => \&gi], [size => \&gi], [timesys => \&gi],
                          [time => \&gi], [byo_time => \&gi], [byo_periods => \&gi],
                          [can_time => \&gi], [can_stones => \&gi]);
   my @recv_challenge  = (KGS::Challenge::,
                          [user1 => \@recv_user], [user2 => \@recv_user],
                          [gametype => \&gi],
                          @recv_rules);
   #y @recv_rules      = # defined earlier
   my @recv_game       = (KGS::Game::,
                          [channel => \&gw],
                          [type  => \&gi],
                          [user1 => \@recv_user], # white
                          [user2 => \@recv_user], # black
                          [user3 => \@recv_user], # owner
                          [size => \&gi],
                          [handicap => \&gi], # < 0 not fully setup
                          [komi => \&gi2],
                          [moves => \&gw],
                          [flags => \&gw],
                          [observers => \&gi],
                          [saved => \&gf],
                          [notes => \&gz, sub { !exists $_[0]{notes} && $_[0]{handicap} < 0 }]);
   my @recv_room_obs   = ([name => \&gs50], [channel => \&gw], [flags => \&gi], [users => \&gi]);
   my @recv_room       = (KGS::Room::, [channel => \&gw],
                          [flags => \&gb], [group => \&gb], [users => \&gw], [games => \&gw], [name => \&gz]);
   my @recv_score      = (KGS::Score::, [score => \&gi4], [territory => \&gi], [captures => \&gi],
                          [i3 => \&gi], [f2 => \&gi4], [komi => \&gi4], [i4 => \&gi]);
   my @recv_stats      = (KGS::Stats::, [ver_major => \&gw], [ver_minor => \&gw], [ver_micro => \&gw],
                          [boot_time => \&gl],
                          [users_cur => \&gi], [users_max => \&gi], [users_lim => \&gi],
                          [accts_cur => \&gi], [accts_max => \&gi], [unknown1 => \&gi],
                          [work_max => \&gi],
                          [rooms_cur => \&gi], [rooms_max => \&gi], [rooms_lim => \&gi],
                          [games_cur => \&gi], [games_max => \&gi], [games_lim => \&gi],
                          [results_cur => \&gi], [results_max => \&gi], [unknown2 => \&gi],
                          [params_cur => \&gi], [params_max => \&gi],
                          [bytes_in => \&gl], [packets_in => \&gl], [bytes_out => \&gl], [packets_out => \&gl]);

   sub guard_repeat {
      1;
   }

   gen_recv login => 0x0001, [result => "ok"], [success => => 1], @recv_user; # more following?
   gen_recv login => 0x0002, [result => "guest"], [success => => 1], @recv_user; # more following?
   gen_recv login => 0x0003, [result => "error3"], @recv_user; # more following?
   gen_recv login => 0x0004, [result => "wrong password"], @recv_user; # more following?
   gen_recv login => 0x0005, [result => "unknown user"], @recv_user; # more following?
   gen_recv login => 0x0006, [result => "user exists"], @recv_user; # more following?

   gen_recv msg_chat   => 0x0013, [user1 => \&ga12], [user2 => \&ga12], [message => \&gz];
   gen_recv stats      => 0x0015, @recv_stats;
   gen_recv idle_warn  => 0x0016; # idle warning, autologout soon
   gen_recv ping       => 0x001d;

   #                        0307 # obsolete room listing (.57 and below)
   gen_recv priv_room  => 0x0310, [name => \&gz]; # permission denied

   # room messages
   gen_recv join_room  => 0x4300, [users => \@recv_user, \&guard_repeat];
   gen_recv msg_room   => 0x4301, [name => \&ga12], [message => \&gz];
   gen_recv part_room  => 0x4302, @recv_user;
   gen_recv del_room   => 0x4303;
   gen_recv upd_games  => 0x4304, [games => \@recv_game, \&guard_repeat];
   gen_recv desc_room  => 0x4319, [owner => \&ga12], [description => \&gz];

   gen_recv chal_defaults => 0x0411, @recv_chal_def;
   gen_recv timewarning_default => 0x001b, [time => \&gw]; # WILD guess

   # hmm.. game-channel messages(?)
   gen_recv upd_chal      => 0x4400, @recv_challenge;
   # total bullshit
   # data => "020000000013000d0002de0300001e000000030061006e0079006f006e0065002000320035006b0020002d00200036006400200074006f006f0020006600610073007400200074006f00200073006c0065006500700020002d0020006f006e006c007900200077006900740068002000720061006e006b00",
   gen_recv upd_game      => 0x4401, [game => \@recv_game];
   gen_recv del_game      => 0x4402;
   gen_recv upd_observers => 0x4403, [users => \@recv_user, \&guard_repeat];
   gen_recv del_observer  => 0x4404, [name => \&ga12];
   gen_recv set_tree      => 0x4405, [tree => \&gtree];
   gen_recv upd_tree      => 0x4406, [tree => \&gtree];
   gen_recv set_node      => 0x4407, [node => \&gi];
   gen_recv repeat_board  => 0x4409;
   gen_recv final_result  => 0x440b, [blackscore => \@recv_score], [whitescore => \@recv_score];
   gen_recv req_undo      => 0x440e;
   gen_recv resign_game   => 0x4410, [player => \&gb]; # 0 is black
   gen_recv set_teacher   => 0x441a, [name => \&ga12];
   gen_recv owner_left    => 0x441d; #d#
   gen_recv teacher_left  => 0x441e; #d#
   gen_recv unknown4422   => 0x4422, [name1 => \&pa12], [name2 => \&pa12], [comment => "change teacher? something to do with editing?"];
   gen_recv req_result    => 0x4433; # # @recv_result
   gen_recv unknown4434   => 0x4434, [b1 => \&gb], [comment => "?? !demonstration game"];

   # class ?= {main, new, clubs, lessons, tournaments, social}
   # flags 0x10 private
   gen_recv upd_rooms     => 0x0318, [rooms => \@recv_room, \&guard_repeat];
   #     int1 => 8,
   #     flags1 => 83886131,
   #     string1 => "MIG team",
   #     type => "unknown318"

   gen_recv upd_game2     => 0x041c, [channel_junk => \&gw], [game => @recv_game];

   # loc 0 => chat(?) loc 1 => gameinfo?, loc 2 => game result (morwe data)
   gen_recv upd_user      => 0x0202, [location => \&gi], [user => \@recv_user],
                                     [lotsofinfo => &gany, sub { !exists $_[0]{lotsofinfo} && $_[0]{location} == 2 }];
   #         data => "0000000068696b6172697368696e0000690a0000",

}

sub decode_msg {
   my ($type, $data) = unpack "v a*", $_[0];

   if ($recvmsgs->{$type}) {
      ($type, $data) = decode_template $recvmsgs->{$type}, $data;
      $type->{data} = unpack "H*", $data if length $data;
      $type;
   } else {
      {
         type => (sprintf "%04x", $type),
         data => (unpack "H*", $data),
      };
   }
}

my $CLIENTVERSION = 4; # 4 == 3 + server zlib compressed
my $SERVERVERSION = 3; # since quite some time now

=item format_msg $name, %args

Returns a formatted message of the given type and payload (which is just concatenated).

=item msg 

Format a message and send it to the server.

=cut

# EVERY message consists of
# pw 0		possibly unused (never seen != 0 (misguided 32 bit ext.??))
# pw length	length of packet in bytes including initial header
# pw type	the packet type
# ?? payload	packet-dependent extra data
#
# if type has bit 14 set, a pw for the channel is following

sub format_msg($@) {
   my ($self, $type, %arg) = @_;
   my $msg = "";

   for (@{$sendmsgs->{$type}}) {
      my ($name, $cvt, $default) = @$_;
      if (defined $arg{$name}) {
         $msg .= $cvt->($arg{$name});
      } elsif (defined $default) {
         $msg .= $cvt->($default);
      } else {
         croak "message type $sendmsgs->{$type}[0][2] requires argument $name\n";
      }
   }
   (pack "vv", 0, (length $msg) + 4) . $msg;
}

sub xdump {
   printf "<%d>\n", length $_[0];
   open my $xt, "|xtype";
   print $xt $_[0];
}

sub msg($@) {
   my $self = shift;
   print ">MSG ", (join ":", @_), "\n";
   my $msg = $self->format_msg(@_);
   xdump $msg;

   if ($self->{sock}) {
      syswrite $self->{sock}, $msg
         or croak "msg: $!";
   } else {
      warn "KGS::Protocol::msg() called without a socket connection\n";
   }
}

=item handshake $fh

Do the initial handshaking with the server on the given socket and make it
the current one. This will initialize the communications stack and set the
socket to be used to sending messages.

"Sockets" should not be anywehre near this module, so this will change.

=cut

sub handshake {
   my ($self, $fh) = @_;

   $self->{sock} = $fh;
   
   syswrite $fh, chr $CLIENTVERSION
      or croak "initial server handshake send: $!";

   sysread $fh, my $buf, 1
      or croak "initial server handshake recv: $!\n";

   $buf eq chr $SERVERVERSION
      or croak "initial server handshake: server version ".(ord $buf)." unsupported";

   $self->{zstream} = inflateInit;
}

=item login $name, $pass[, $locale[, $vm]]

=cut

sub login {
   my ($self, $name, $pass, $locale, $vm) = @_;
   $self->{name} = $name;
   $self->msg(login =>
      login    => $name,
      password => (hash_pw $pass),
      locale   => $locale,
      guest    => $pass eq "",
      javavm   => $vm,
   );
}

=item disconnect

Disconnect the socket.

=cut

sub disconnect {
   my $self = shift;

   $self->{sock}->close if $self->{sock};
}

=item feed_data $data

Feed new data from the server into this object. This might or might not
result in messages being dispatched.

=cut

sub feed_data {
   my $self = shift;

   $NOW = Time::HiRes::time;

   my ($data, $status) = $self->{zstream}->inflate($_[0]);
   $status == Z_OK or croak "inflate: status is $status";

   $self->{rbuf} .= $data;

   while (($self->{rlen} || 4) <= length $self->{rbuf}) {
      my $data = substr $self->{rbuf}, 0, $self->{rlen} || 4, "";
      if (delete $self->{rlen}) {
         $self->_inject(decode_msg $data);
      } else {
         $self->{rlen} = (unpack "xxv", $data) - 4;
      }
   }
}

sub register {
   my ($self, $obj, @types) = @_;

   $self->{cb}{$_}{$obj} = $obj for @types || "any";
}

sub unregister {
   my ($self, $obj, @types) = @_;

   delete $self->{cb}{$_}{$obj} for @types || "any";
}

sub _inject {
   my ($self, $msg) = @_;

   for my $type ($self->{type}, "any") {
      for my $obj (values %{$self->{cb}{$type} || {}}) {
         $obj->inject($msg);
      }
   }
}

package KGS::User;

sub is_guest	{ $_[0]{flags} & 0x0001 }
sub is_admin	{ $_[0]{flags} & 0x0002 }
sub is_active	{ $_[0]{flags} & 0x0008 } # logged in(?)
sub is_away	{ $_[0]{flags} & 0x0010 }
sub is_idle	{ $_[0]{flags} & 0x0020 }
sub is_playing	{ $_[0]{flags} & 0x0080 }
sub is_reliable	{ $_[0]{flags} & 0x0100 } # reliable ranking?
sub is_ranked	{ $_[0]{flags} & 0x0200 }
sub has_pic	{ $_[0]{flags} & 0x1000 }
sub email_priv	{ $_[0]{flags} & 0x2000 }

sub _rank {
   $_[0] <= 0
      ? ""
      : $_[0] <= 30
         ? (31-$_[0]) . "k"
         : $_[0] <= 37
              ? ($_[0] - 30) . "d"
              : ($_[0] - 37) . "p";
}

sub rank {
   $_[0]->is_ranked && 0
      ? "-"
      : _rank($_[0]{flags} >> 24) . ($_[0]->is_reliable ? "" : "?");
}

sub as_string {
   sprintf "%s [%s]", $_[0]{name}, $_[0]->rank;
}

package KGS::Game;

sub is_scored    { $_[0]{flags} & 0x1 }
sub is_adjourned { $_[0]{flags} & 0x2 }
# there is more going on, one bit 0x4 and above(?)

sub is_valid     { $_[0]{handicap} >= 0 }

sub moves {
   $_[0]->is_scored
      ? 0
      : $_[0]{moves};
}

sub type {
   (qw(D T F R O C))[$_[0]{type} % 5];
}

sub privacy {
   ("", "-lecture", "-private")[$_[0]{type} / 5];
}

sub user0 {
   length $_[0]{user3}{name}
      ? $_[0]{user3}->as_string
      : $_[0]{user2}->as_string;
}

sub user1 {
   $_[0]{user3}{name} ?
      $_[0]{user1}{name} eq $_[0]{user2}{name}
         ? "" : "(".$_[0]{user2}->as_string." - ".$_[0]{user1}->as_string.")"
      : $_[0]{user1}->as_string;
}

sub size {
   $_[0]{size};
}

# convinience

sub rules {
   my $rules = sprintf "%dx%d", ($_[0]{size})x2;
   $rules .= " H$_[0]{handicap}"
           . ($_[0]{komi} < 0 ? $_[0]{komi} : "+$_[0]{komi}") if $_[0]->is_valid;
   $rules;
}

sub notes {
   $_[0]->is_valid
      ? $_[0]->is_scored
         ? "[finis]"
         : sprintf "Mv %d Ob %d", $_[0]{moves}, $_[0]{observers}
   : $_[0]{notes};
}

sub uri {
   my $game = shift;

   next if $status & 0x80; # inPlay
   next if $scoreval == 16387; # finished

   next if int($flags / 5) == 2;

   # low 4 bits are handicap
   $flags = ($flags >> 4) & 15;

   my $type    = $self->{type};
   my $privacy = $self->{privacy};

   $revision = (($f1 >> 16) & 0x00ff) | (($f2 >> 8) & 0xff00);

   my $p1 = $self->{user1}{name};
   my $p2 = $self->{user2}{name};
   my $p3 = $self->{user3}{name};

   my ($year, $month, $day) = (gmtime $date * 0.001)[5,4,3];

   my $playerstring = $p3 ? $p3 : $p1 eq $p2 ? $p1 : "$p2-$p1";
   my $revisionstring = $revision ? "-" . ($revision+1) : "";

   my $url = sprintf "/games/%d/%d/%d/%s%s.sgf",
                     $year + 1900, $month + 1, $day,
                     $playerstring, $revisionstring;

   $url;
}

package KGS::Room;

sub is_default { $_[0]{flags} & 0x04 }
sub is_private { $_[0]{flags} & 0x10 }

package KGS::Challenge;

package KGS::Score;

package KGS::Stats;

1;



